// FBX1.cpp : Defines the entry point for the console application.
//

#include <fbxsdk.h>
#include <assert.h>

FbxManager *gFbxManager = NULL;

//
//	Routines to create and destroy a single SDK manager for the
//	program
//
FbxManager *globalSdkMgr()
{
	if( gFbxManager == NULL )
	{
		gFbxManager = FbxManager::Create();
		assert( gFbxManager != NULL );
	}
	return gFbxManager;
}

void globalMgrDestroy()
{
	gFbxManager->Destroy();
	gFbxManager = NULL;
}

// Create a directional camera with target
// 
FbxNode* CreateCamera(FbxScene* pScene, FbxVector4 vecFrom, FbxVector4 vecTo )
{
    FbxCamera* lCamera = FbxCamera::Create(pScene,"Camera");

    // Modify some camera default settings.
    lCamera->SetApertureMode(FbxCamera::eVertical);
    lCamera->SetApertureWidth(0.816);
    lCamera->SetApertureHeight(0.612);
    lCamera->SetSqueezeRatio(0.5);
    
    FbxNode* lNodeCamera = FbxNode::Create(pScene,"CameraNode");
    lNodeCamera->LclTranslation.Set(vecFrom);
    lNodeCamera->SetNodeAttribute(lCamera);
   
    FbxMarker* lMarker = FbxMarker::Create(pScene,"TestCameraTarget");

    FbxNode* lNodeTarget = FbxNode::Create(pScene,"TestCameraTarget");
    lNodeTarget->LclTranslation.Set(vecTo);
    lNodeTarget->SetNodeAttribute(lMarker);

    lNodeCamera->SetTarget(lNodeTarget);

    return lNodeCamera;
}

// Create a light.
//
FbxNode* CreateLight(FbxScene* pScene, FbxDouble3 position)
{
    FbxString lLightName;
    lLightName = "PointLight";

    FbxLight* lFbxLight = FbxLight::Create(pScene, lLightName.Buffer());

    lFbxLight->LightType.Set(FbxLight::ePoint);

    // Light Color...
    //
    FbxDouble3 lColor;
    lColor[0] = 1;
    lColor[1] = 1;
    lColor[2] = 1;
    lFbxLight->Color.Set(lColor);

    // Light Intensity...
    //
    lFbxLight->Intensity.Set(100.0);

    FbxNode* lNode = FbxNode::Create(pScene, lLightName+"Node");

    // set the node attribute
    lNode->SetNodeAttribute(lFbxLight);
    lNode->LclTranslation.Set(position);

    return lNode;
}

// Create user defined properties
//
void CreateUserProperties(FbxNode *pNode, FbxObject *pObj) 
{
    // Now we create the user properties 
    FbxProperty custPropBool = FbxProperty::Create(pNode, FbxBoolDT, "ADN_Bool", "ADN Sample Bool Property");
    FbxProperty custPropInt = FbxProperty::Create(pNode, FbxIntDT, "ADN_Integer", "ADN Sample Integer Property");
    FbxProperty custPropStrList = FbxProperty::Create(pNode, FbxStringListDT, "ADN_StringList", "");

    custPropBool.ConnectSrcObject(pObj, FbxConnection::eReference);
    custPropInt.ConnectSrcObject(pObj, FbxConnection::eReference);
    custPropStrList.ConnectSrcObject(pObj, FbxConnection::eReference);

    /* 
    NOTE: The properties labels exists only while the property object is in memory.
    The label is not saved in the FBX file. When loading properties from the FBX file
    it will take the same value as the property name.
    */

    // we now fill the properties. All the properties are user properties so we set the
    // correct flag
    custPropBool.ModifyFlag(FbxPropertyAttr::eUser, true);
    custPropInt.ModifyFlag(FbxPropertyAttr::eUser, true);
    custPropStrList.ModifyFlag(FbxPropertyAttr::eUser, true);

    // we set the default values
    custPropBool.Set(false);
    custPropInt.Set(0);

    // and some limits
    custPropInt.SetLimits(0, 10);

    // add elements to the list
    custPropStrList.AddEnumValue("one");
    custPropStrList.AddEnumValue("two");
    custPropStrList.AddEnumValue("three");
    custPropStrList.AddEnumValue("Four");
    custPropStrList.InsertEnumValue(0, "zero");
}

void PrintLevel(int level, bool bPlus = false)
{
	// Level indentation
	for( int i = 0; i < level; i++ )
	{
        if ((i+1 == level) && (bPlus))
		    FBXSDK_printf( "+-" );
        else
            FBXSDK_printf( "  " );
	}
}

// Iterate the scene, and for each node and 
// object look at the connections
//
void GetConnections( FbxNode *currentNode, int level = 0 )
{

	//	Print out the node name
	//
    PrintLevel(level, true);
    FBXSDK_printf("Node %s \n", currentNode->GetName());

    FbxObject* currentObj = currentNode->GetNodeAttribute();
    if (currentObj)
    {
        int numSrcProps = currentObj->GetSrcPropertyCount();
        for (int j = 0; j < numSrcProps; j++)
        {
		    FbxProperty propSrc = currentObj->GetSrcProperty(j);
            PrintLevel(level, false);
            FBXSDK_printf(" Src Prop: %s\n", propSrc.GetNameAsCStr());
        }
        int numDstProps = currentObj->GetDstPropertyCount();
        for (int j = 0; j < numDstProps; j++)
        {
		    FbxProperty propDst = currentObj->GetDstProperty(j);
            PrintLevel(level, false);
            FBXSDK_printf(" Dst Prop: %s\n", propDst.GetNameAsCStr());
        }
    }

	//	Get and Print out the Source Objects
	//
    int numSrcObjs = currentNode->GetSrcObjectCount();
	for (int i = 0; i < numSrcObjs; i++)
    {
		FbxObject *objSrc = currentNode->GetSrcObject(i);
        PrintLevel(level, false);
        FBXSDK_printf(" Src Obj: %s\n", objSrc->GetName());
	}

	//	Get and Print out the Destination Objects
	//
    int numDstObjs = currentNode->GetDstObjectCount();
	for (int i = 0; i < numDstObjs; i++)
    {
		FbxObject *objDst = currentNode->GetDstObject(i);
        PrintLevel(level, false);
        FBXSDK_printf(" Dst Obj: %s\n", objDst->GetName());
	}

	// Recurse on the node children
	int numKids = currentNode->GetChildCount();
	for (int k = 0; k < numKids; k++)
    {
		FbxNode *child = currentNode->GetChild(k);
		GetConnections(child, level+1);
	}
}

// Create a cube
//
FbxNode *CreateCube()
{
	//	1. Define cube vertices, face normals, and face definitions
	//
	const int numFaces = 6;
	const int vertsPerFace = 4;

	FbxVector4 points[] = {
		FbxVector4( -50, -50, 50 ),
		FbxVector4( 50, -50, 50 ),
		FbxVector4( 50, 50, 50 ),
		FbxVector4( -50, 50, 50 ),
		FbxVector4( -50, -50, -50 ),
		FbxVector4( 50, -50, -50 ),
		FbxVector4( 50, 50, -50 ),
		FbxVector4( -50, 50, -50 ) };

	FbxVector4 normals[] = {
		FbxVector4( 0, 0, 1 ),
		FbxVector4( 1, 0, 0 ),
		FbxVector4( 0, 0, -1 ),
		FbxVector4( -1, 0, 0 ),
		FbxVector4( 0, 1, 0 ),
		FbxVector4( 0, -1, 0 ) };

	int faceVerts[] = { 0, 1, 2, 3, 
					 1, 5, 6, 2, 
					 5, 4, 7, 6, 
					 4, 0, 3, 7, 
					 3, 2, 6, 7, 
					 4, 5, 1, 0 };

	//	2. Create the mesh object (FbxMesh)
	//
	FbxMesh *myMesh = NULL;
	myMesh = FbxMesh::Create( globalSdkMgr(), "myCube" );

	//	3. Initialize the control point array
	//
	//	Use FbxMesh::InitControlPoints(), FbxMesh::GetControlPoints()
	//	To split normals at edges, we will add separate control points 
	//	for each face
	myMesh->InitControlPoints( numFaces*vertsPerFace );
	FbxVector4* cpArray = myMesh->GetControlPoints();

	//	4. Add vertices to the array, and define polygon faces
	//
	//	FbxMesh::BeginPolygon(), FbxMesh::EndPolygon() begin and end face
	//	definitions.  In between, FbxMesh::AddPolygon() adds indexes to 
	//	control points.
	//
	//	Also fill in the control point array here - there should be 
	//	24 control points, in order to separate per-face normals at the
	//	vertices.
	//
	for( int f = 0; f < numFaces; f++ )
	{
		myMesh->BeginPolygon();
		for( int v = 0; v < vertsPerFace; v++ )
		{
			int vIndex = vertsPerFace*f+v;

			myMesh->AddPolygon( vIndex );
			cpArray[vIndex] = points[faceVerts[vIndex]];
		}
		myMesh->EndPolygon();
	}

	//	5. Create a layer to store normals
	//
	FbxLayer* layer0 = myMesh->GetLayer(0);
	if( layer0 == NULL )
	{
		myMesh->CreateLayer();
		layer0 = myMesh->GetLayer(0);
	}

	//	6. Add normals to the layer
	// Use FbxLayerElementNormal::SetMappingMode() and SetReferenceMode().
	//
	FbxLayerElementNormal* normalLayer = FbxLayerElementNormal::Create(myMesh, "");
	normalLayer->SetMappingMode( FbxLayerElement::eByControlPoint );
	normalLayer->SetReferenceMode( FbxLayerElement::eDirect );
	// Associate the normal layer element data with layer 0
	layer0->SetNormals( normalLayer );

	FbxLayerElementArrayTemplate<FbxVector4>& normalArray = normalLayer->GetDirectArray();
	// fill in the normals.....
	for( int f = 0; f < numFaces; f++ )
	{
		for( int v = 0; v < vertsPerFace; v++ )
		{
			normalArray.Add( normals[f] );
		}
	}

	//	7. Create a node for the mesh, and set its node attribute to the mesh data
	//
	FbxNode *myNode = FbxNode::Create( globalSdkMgr(), "myNode"  );
	myNode->SetNodeAttribute( myMesh );

	return myNode;
}

// Create materials and assign to a cube
//
void MaterialCube(FbxNode* pCube, FbxScene* scene)
{   
    FbxMesh* mesh = pCube->GetMesh();
    FbxGeometryElementMaterial* mat = mesh->CreateElementMaterial();
    mat->SetMappingMode(FbxLayerElement::eByPolygon);
    mat->SetReferenceMode(FbxGeometryElement::eIndexToDirect);

    int count = mesh->GetPolygonCount();

    mat->GetIndexArray().SetCount(count);
    for (int i = 0; i < count; i++ )
    {
        FbxString lMaterialName = "TestMaterial";
        FbxString lShadingModelName = i%2==0 ? "Lambert" : "Phong";
        lMaterialName += i;
        FbxDouble3 lEmissive(0.0, 0.0, 0.0);
        FbxDouble3 lAmbient(1.0, 0.0, 0.0);
        FbxDouble3 lColor;
        FbxSurfaceLambert *lMaterial = FbxSurfaceLambert::Create(scene, lMaterialName.Buffer());

        lMaterial->Emissive.Set(lEmissive);
        lMaterial->Ambient.Set(lAmbient);
        lColor = FbxDouble3(i > 2   ? 1.0 : 0.0, 
            i > 0 && i < 4 ? 1.0 : 0.0, 
            i % 2   ? 0.0 : 1.0);
        lMaterial->Diffuse.Set(lColor);

        lMaterial->TransparencyFactor.Set(0.0);
        lMaterial->ShadingModel.Set(lShadingModelName);

        mat->GetIndexArray().SetAt(i,i);

        pCube->AddMaterial(lMaterial);

    }
}


// Animate the cube along one axis
//
void AnimateCube(FbxNode* pCube, FbxScene* scene)
{
	FbxAnimStack* animStack = FbxAnimStack::Create(scene, "Cube Animation Stack");
	FbxAnimLayer* animLayer = FbxAnimLayer::Create(scene, "Base Layer");
	animStack->AddMember(animLayer);

    FbxTime lTime;
    int lKeyIndex = 0;

    FbxAnimCurve* acurve = pCube->LclTranslation.GetCurve<FbxAnimCurve>(animLayer, FBXSDK_CURVENODE_COMPONENT_Z, true);
    if (acurve)
    {
        acurve->KeyModifyBegin();

        lTime.SetSecondDouble(0.0);
        lKeyIndex = acurve->KeyAdd(lTime);
        acurve->KeySet(lKeyIndex, lTime, 0.0, FbxAnimCurveDef::eInterpolationLinear);

        lTime.SetSecondDouble(2.0);
        lKeyIndex = acurve->KeyAdd(lTime);
        acurve->KeySet(lKeyIndex, lTime, 300.0, FbxAnimCurveDef::eInterpolationLinear);

        acurve->KeyModifyEnd();
    }

}

// OPTIONAL: Create a sphere. (Helper function)
//
FbxNode* CreateNurbsSphere(FbxScene* pScene, char* pName)
{
    FbxNurbs* lNurbs = FbxNurbs::Create(pScene,pName);

    // Set nurbs properties.
    lNurbs->SetOrder(4, 4);
    lNurbs->SetStep(2, 2);
    lNurbs->InitControlPoints(8, FbxNurbs::ePeriodic, 7, FbxNurbs::eOpen);

    double lUKnotVector[] = { -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0 };
    memcpy(lNurbs->GetUKnotVector(), lUKnotVector, lNurbs->GetUKnotCount()*sizeof(double));

    double lVKnotVector[] = { 0.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0, 4.0, 4.0 };
    memcpy(lNurbs->GetVKnotVector(), lVKnotVector, lNurbs->GetVKnotCount()*sizeof(double));

    FbxVector4* lVector4 = lNurbs->GetControlPoints();
    int i, j;
    double lScale = 20.0;
    double lPi = 3.14159;
    double lYAngle[] = { 90.0, 90.0, 52.0, 0.0, -52.0, -90.0, -90.0 };
    double lRadius[] = { 0.0, 0.283, 0.872, 1.226, 0.872, 0.283, 0.0}; 

    for (i = 0; i < 7; i++) 
    {
        for (j = 0; j < 8; j++)
        {
            double lX = lScale * lRadius[i] * cos(lPi/4*j);
            double lY = lScale * sin(2*lPi/360*lYAngle[i]);
            double lZ = lScale * lRadius[i] * sin(lPi/4*j);
            double lWeight = 1.0;

            lVector4[8*i + j].Set(lX, lY, lZ, lWeight);
        }
    }


    FbxNode* lNode = FbxNode::Create(pScene,pName);

    lNode->SetNodeAttribute(lNurbs);

    return lNode;
}

// OPTIONAL: Map texture over sphere. (Helper function)
//
void MapSphereTexture(FbxScene* pScene, FbxNode* pNurbs)
{
    FbxFileTexture* lTexture = FbxFileTexture::Create(pScene,"eye.jpg");

    // The texture won't be displayed if node shading mode isn't set to FbxNode::eTextureShading.
    pNurbs->SetShadingMode(FbxNode::eTextureShading);

    // Set texture properties.
    lTexture->SetFileName("eye.jpg"); // Resource file is in current directory.
    lTexture->SetTextureUse(FbxTexture::eStandard);
    lTexture->SetMappingType(FbxTexture::eSpherical);
    lTexture->SetMaterialUse(FbxFileTexture::eModelMaterial);
    lTexture->SetSwapUV(false);
    lTexture->SetTranslation(0.0, 0.0);
    lTexture->SetScale(1.0, 2.0);
    lTexture->SetRotation(45.0, 0.0);

    // we connect the texture to the material DiffuseColor property
    FbxSurfacePhong* lMaterial = pNurbs->GetSrcObject<FbxSurfacePhong>(0);
    if (lMaterial)
        lMaterial->Diffuse.ConnectSrcObject(lTexture);

}

// OPTIONAL: Map material over sphere. (Helper function)
//
void MapSphereMaterial(FbxScene* pScene, FbxNode* pNurbs)
{
    FbxSurfacePhong* lMaterial = FbxSurfacePhong::Create(pScene,"scene02");
    FbxDouble3 lBlue(0.05, 0.05, 0.05);
    FbxDouble3 lBlack(0.0, 0.0, 0.0);

    lMaterial->Emissive.Set(lBlue);
    lMaterial->Ambient.Set(lBlack);
    lMaterial->Specular.Set(lBlack);
    lMaterial->TransparencyFactor.Set(0.0);
    lMaterial->Shininess.Set(0.0);
    lMaterial->ReflectionFactor.Set(0.0);

    // Create GeometryElementMaterial
    FbxNurbs*						lNurbs					= pNurbs->GetNurbs();
    FbxGeometryElementMaterial*	lGeometryElementMaterial	= lNurbs->GetElementMaterial( 0);

    if (!lGeometryElementMaterial)
    {
        lGeometryElementMaterial = lNurbs->CreateElementMaterial();
    }

    // The material is mapped to the whole Nurbs
    lGeometryElementMaterial->SetMappingMode(FbxGeometryElement::eAllSame);

    // And the material is avalible in the Direct array
    lGeometryElementMaterial->SetReferenceMode(FbxGeometryElement::eDirect);
    pNurbs->AddMaterial(lMaterial);
}

// OPTIONAL: Create and assign texture to sphere with proper UV mapping
//
void CreateTextureSphere(FbxScene *scene)
{
    FbxNode* lNurbs = CreateNurbsSphere(scene, "TestSphere");

   // MapShapesOnNurbs(scene, lNurbs);
    MapSphereMaterial(scene, lNurbs);
    MapSphereTexture(scene, lNurbs);

    // Build the node tree.
    FbxNode* lRootNode = scene->GetRootNode();
    lRootNode->AddChild(lNurbs);
    lNurbs->LclTranslation.Set(FbxVector4(0.0, 0.0, 200.0));
}

int findAsciiFormat (FbxManager *mgr)
{
    // Write in fall back format in less no ASCII format found
    int fileFormat = mgr->GetIOPluginRegistry()->GetNativeWriterFormat();

    //Try to export in ASCII if possible
    int lFormatIndex, lFormatCount = mgr->GetIOPluginRegistry()->GetWriterFormatCount();

    for (lFormatIndex=0; lFormatIndex<lFormatCount; lFormatIndex++)
    {
        if (mgr->GetIOPluginRegistry()->WriterIsFBX(lFormatIndex))
        {
            FbxString lDesc =mgr->GetIOPluginRegistry()->GetWriterFormatDescription(lFormatIndex);
            char *lASCII = "ascii";
            if (lDesc.Find(lASCII)>=0)
            {
                fileFormat = lFormatIndex;
                break;
            }
        }
    }

    return fileFormat;
    
}

void ExportFbxScene( FbxManager *mgr, FbxScene *scene, char *pFileName )
{
	FbxIOPluginRegistry *pluginRegistry = mgr->GetIOPluginRegistry();
	assert( pluginRegistry != NULL );

	//	create an exporter
	//
	FbxExporter *exporter = FbxExporter::Create( mgr, "ADN_Exporter" );

    // Initize settings
    FbxIOSettings * ios = FbxIOSettings::Create(mgr, IOSROOT );
    //ios->SetBoolProp(EXP_FBX_ANIMATION, true); // Do not need animation in this example
    mgr->SetIOSettings(ios);


    int fileFormat = findAsciiFormat(mgr);

    // Initialize the exporter by providing a filename.
    if(exporter->Initialize(pFileName, fileFormat, mgr->GetIOSettings()) == false)
    {
        FBXSDK_printf("Call to FbxExporter::Initialize() failed.\n");
        FBXSDK_printf("Error returned: %s\n\n", exporter->GetLastErrorString());
        return;
    }

    if (!exporter->Export(scene))
        FBXSDK_printf("Call to FbxExporter::Export() failed.\n");

    return;
}

int main(int /*argc*/, char* /*argv[]*/)
{
    // Lab for Unit 3
	//	create an SDK manager
	//
	FbxManager *fbxSdkManager = globalSdkMgr();

	//	create a FbxScene
	//
	FbxScene *scene = FbxScene::Create( fbxSdkManager, "ADN_Scene" );
    
    // Get the root node
    //
    FbxNode *root = scene->GetRootNode();

    // Add a Directional camera to the root node of the scene
    // It should be postioned at 850.0, 700.0, 200.0, and be directed at 0.0, 0.0, 0.0
    //
    FbxNode* nodeCamera = CreateCamera(scene, FbxVector4(850.0, 700.0, 200.0), FbxVector4(0.0, 0.0, 0.0));
    root->AddChild(nodeCamera);

    // Add a light to the root node of the scene. Position it at 1000,1000,1000.
    //
    FbxNode* nodeLight = CreateLight(scene, FbxDouble3(1000, 1000, 1000));
    root->AddChild(nodeLight);
    
    // Create some user defined properties
    // Add these properties to the light node, 
    // and also reference them with a connection to the camera object
    //
    CreateUserProperties(nodeLight, nodeCamera->GetNodeAttribute());

    // Try to find the properties.
    // First on the light node.
    FbxProperty propFromLightNode = nodeLight->FindProperty("ADN_StringList", false);
    if (propFromLightNode.IsValid())
        FBXSDK_printf("\nFound Light Node Prop: %s", propFromLightNode.GetNameAsCStr());
    // Then on the Camera object (NodeAttribute base). Here it will fail because the property is
    // not directly contained by the object or it's node. Instead a connection reference is made to
    // it. It will be found through the property connections APIs, though.
    FbxProperty propFromCamera = nodeCamera->GetNodeAttribute()->FindProperty("ADN_StringList", false);
    if (propFromCamera.IsValid())
        FBXSDK_printf("\nFound Camera Prop: %s", propFromCamera.GetNameAsCStr());

    // Now iterate the scene's nodes and associated objects and get the connection infomration.
    //
    FBXSDK_printf("\n\n\nObject Connections\n");
    GetConnections(root);

    // Lab for Unit 4
	//	create a cube and add it to the scene
	//
	FbxNode *nodeCube = CreateCube();
	root->AddChild( nodeCube );

    // Add meterials to the cube. A different color material for each side.
    //
    MaterialCube(nodeCube, scene);
    
    // Create a sphere and add a texture to it.
    //
    CreateTextureSphere(scene);

	//	animate the cube to move along one axis
	//
	AnimateCube(nodeCube, scene);

	//	save the scene to ASCII FBX file
	//
	ExportFbxScene( fbxSdkManager, scene, "output.fbx" );

	//	clean up memory for the manager
	//
	globalMgrDestroy();

    // Add a puase for ease of watching outputs during test runs
    scanf("pause");

	return 0;
}

